import os
import sys
import platform
import logging
import signal
import subprocess
import threading
import time
import glob
import shutil
import tempfile
import re
# Hack to make the sdist work
try:
    import jetpack.messaging
except ImportError:
    pass
from healthcheck.check import Check, CheckError

logger = logging.getLogger('healthcheck')

SHUTDOWN_MSG = '''
HealthCheck Daemon has detected a system problem
Error Message: %s
The system will be shutdown in %s minutes
To delay the shutdown, execute the command `keepalive N` where N represents the number of minutes to delay system termination.
'''
STAY_OF_EXECUTION_MSG = '''HEALTHCHECK: system termination has been manually delayed
        The system will be shutdown in %s minutes
To delay the shutdown again, execute the command `jetpack keepalive N` where N represents the number of minutes to delay system termination.'''

class HealthService:

    def _send_log_message(self, message, level='info'):
        '''
        Send a log message to CycleCloud
        '''
        try:
            jetpack.messaging.log(message, level)
        except:
            # Failing to send the message shouldn't break things for us
            logger.error('Could not send message to CycleCloud')

    def make_parent_dir(self, path):
        parent_dir = os.path.dirname(path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)

    def check_pid_in_use(self, pid):
        try:
            os.kill(pid, 0)
        except OSError:
            pid_in_use = False
        else:
            pid_in_use = True
            logger.fatal('A process id, %s, appears to be in use for this program.\nIf this is incorrect, delete the pidfile at %s. bailing out', 
                        pid, self.pidfile)
            os.sys.exit(0)

    def check_and_write_pid(self):
        if os.path.exists(self.pidfile):
            try:
                f = open(self.pidfile, 'r+t')
                pid = f.readline().strip()
                self.check_pid_in_use(int(pid))
                f.close()
            except Exception as err:
                sys.stderr.write('ERROR accessing pidfile %s' % self.pidfile)
                os.sys.exit(1)

        self.make_parent_dir(self.pidfile)
        f = open(self.pidfile, 'w')
        f.write(str(os.getpid()))
        f.close()

    def handle_signals(self):
        def signal_handler(signum, frame):
            logger.info('Signal {0} received, shutting down'.format(signum))
            self.cleanup()
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

    def quit_if_user_disabled(self, stay_of_execution):
        if stay_of_execution == 0:
            logger.info('HealthCheck service has been disabled by the user. Exiting now')
            self.cleanup()
            # shuts down main thread
            if self.shutdown_healthservice_event:
                self.shutdown_healthservice_event.set()
            sys.exit(0)
        else:
            return stay_of_execution

    def check_if_user_disabled_and_quit(self):
        stay_of_execution = self.read_stay_of_execution()
        new_stay_of_execution = self.quit_if_user_disabled(stay_of_execution)
        if self.stay_is_valid(new_stay_of_execution):
            self.stay_of_execution = new_stay_of_execution

    def execute(self, shutdown_healthservice_event=None):
        self.shutdown_healthservice_event = shutdown_healthservice_event
        self.make_parent_dir(self.scriptdir)
        from os import listdir
        # Stay of execution not yet supported on Windows
        if platform.os.name == 'nt':
            logger.warning('Stay of execution not yet supported on Windows')
        else:
            self.check_if_user_disabled_and_quit()
        for f in listdir(self.scriptdir):
            script = os.path.join(self.scriptdir, f)
            if os.path.isdir(script):
                continue
            hc = Check(script, self.timeout, self.sandbox)
            try:
                returncode = hc.run()
                logger.info('Received returncode of %d for script: %s' % (returncode, script))
                if returncode == 254:
                    # Kill the instance if the explicit kill return code is
                    # returned.
                    self.suicide(hc)
                    break
                elif returncode == 0:
                    # We got the successful check code, so do nothing.
                    pass
                else:
                    # Not sure what to do with this return code, so we do
                    # nothing except tell CycleCloud about it.
                    self._send_log_message('HeathCheck script %s failed with \
                            unexpected exit code %d.' % (f, returncode), 'warn')
            except CheckError as e:
                logger.error('Received error running check: %s' % str(e))
                break
        if not self.daemonize:
            # clean up the wall process
            self.cleanup()

    def run(self):
        if platform.os.name == 'posix':
            self.check_and_write_pid()
            self.handle_signals()

        while True:
            logger.info('Beginning execution of scripts in directory {0} at interval {1} minutes'.format(self.scriptdir, self.interval))
            shutdown_healthservice_event = threading.Event()
            self.thread = threading.Timer(float(self.interval) * 60, self.execute, args=[shutdown_healthservice_event])
            self.thread.setDaemon(True)
            self.thread.start()
            while threading.active_count() > 1:
                shutdown_set = shutdown_healthservice_event.wait(1)
                if shutdown_set:
                    sys.exit(0)

    def override_with_cmdline_opts(self, cmdline_options):
        if cmdline_options.__class__ == dict:
            opts_dict = cmdline_options
        else:
            opts_dict = cmdline_options.__dict__
        for opt in opts_dict:
            if opts_dict[opt] != None:
                self.__dict__[opt] = opts_dict[opt]

    def warn_users(self, shutdown_reason):
        if self.platform == 'nt':
            pass
        else:
            shutdown_msg = SHUTDOWN_MSG % (shutdown_reason, self.stay_of_execution)
            self.warn_proc.communicate(shutdown_msg)
            self.warn_proc = self.allocate_shell('wall')

    def announce_stay_of_execution(self):
        msg = STAY_OF_EXECUTION_MSG % self.stay_of_execution
        self.warn_proc.communicate(msg)
        self.warn_proc = self.allocate_shell('wall')

    def convert_time_unit_minutes(self, time_unit):
        if time_unit.endswith('m'):
            return int(time_unit[0:-1])
        elif time_unit.endswith('h'):
            return int(time_unit[0:-1]) * 60
        elif time_unit.endswith('d'):
            return int(time_unit[0:-1]) * 60 * 60
        else:
            # default to hours
            return int(time_unit[0:-1]) * 60

    def read_stay_of_execution(self):
        time_regex = r"""^\d+[mdh]?"""
        stay_of_execution = os.read(self.fifo, 1024).decode()
        if stay_of_execution == '':
            return ''
        else:
            match = re.match(time_regex, stay_of_execution)
            # in case a user sends multiple keepalive messages
            # we only match the first one
            time_unit = match.group()
            if time_unit.startswith('0'):
                return 0
            else:
                return self.convert_time_unit_minutes(time_unit)

    def stay_is_valid(self, stay):
        if stay == '':
            return False
        stay = int(stay)
        if stay > 0:
            return True
        else:
            return False

    def do_stay_of_execution(self):
        execution_manually_stayed = False
        for minute in range(int(self.stay_of_execution)):
            for i in range(60):
                time.sleep(1)
                new_stay_of_execution = self.read_stay_of_execution()
                self.quit_if_user_disabled(new_stay_of_execution)
                if self.stay_is_valid(new_stay_of_execution):
                    self.stay_of_execution = int(new_stay_of_execution)
                    self.announce_stay_of_execution()
                    execution_manually_stayed = True

        if execution_manually_stayed:
            self.announce_stay_of_execution()
            self.do_stay_of_execution()
        
    def suicide(self, health_check, exception=None):
        
        if not exception:
            shutdown_reason = health_check.stderr
            logger.critical('Script {0} exited with return code {1}'.format(health_check.script, health_check.returncode))
            logger.critical('STDERR for script: {0}'.format(shutdown_reason))
            logger.critical('STDOUT for script: {0}'.format(health_check.stdout))
        else:
            shutdown_reason = str(exception)
            logger.critical('Could not execute script {0} due to exception \"{1}\"'.format(health_check.script, str(exception)))
        if shutdown_reason and len(shutdown_reason) > 250:
            check_stderr = shutdown_reason[:100] + '...' + shutdown_reason[-100:]
        else:
            check_stderr = shutdown_reason
        self._send_log_message('HealthCheck script %s requested termination: %s' \
                % (os.path.basename(health_check.script), check_stderr), 'error')
        logger.critical('Shutting down machine in %s minute.' % self.stay_of_execution)
        self.warn_users(shutdown_reason)
        
        try:
            if platform.os.name == 'nt':
                home = os.environ.get('CYCLECLOUD_HOME', r'C:\cycle\jetpack')
                jetpack_command = os.path.join(home, 'bin', 'jetpack.cmd')
            else:
                home = os.environ.get('CYCLECLOUD_HOME', '/opt/cycle/jetpack')
                jetpack_command = os.path.join(home, 'bin', 'jetpack')

            subprocess.check_call("%s shutdown --unhealthy" % jetpack_command, shell=True)
        except KeyboardInterrupt:
            # Keyboard interupt is ok, it means that "jetpack shutdown" was retrying and the user decided to cancel via ctrl-c
            pass
        except Exception as e:
            logger.exception("Unable to run `jetpack shutdown --unhealthy`")

    def allocate_shell(self, cmd):
        p = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True)
        return p

    def cleanup(self):
        # All of the cleanup is currently Linux-oriented
        if platform.os.name == 'nt':
            return
        else:
            if os.path.exists(self.pidfile):
                os.remove(self.pidfile)
            self.cleanup_fifo()
            self.warn_proc.terminate()

    def remove_tmpdirs(self):
        leftover_dirs = glob.glob('/tmp/healthcheck*')
        for dir in leftover_dirs:
            shutil.rmtree(dir)

    def cleanup_fifo(self):
        os.close(self.fifo)
        self.remove_tmpdirs()

    def mkfifo(self):
        self.remove_tmpdirs()
        tmpdir = tempfile.mkdtemp(prefix='/tmp/healthcheck')
        fifo_path = os.path.join(tmpdir,'healthcheck')
        os.mkfifo(fifo_path)
        os.chmod(tmpdir, 0o777)
        os.chmod(fifo_path, 0o777)
        self.fifo = os.open(fifo_path, os.O_RDONLY | os.O_NONBLOCK)

    def _get_tmp_home(self):
        if self.platform == 'nt':
            home = "c:\\windows\\temp"
        else:
            home = "/tmp"
        return home

    def __init__(self, cmdline_options):
        self.platform = platform.os.name
        if 'CYCLECLOUD_HOME' in os.environ:
            jetpack_home = os.environ['CYCLECLOUD_HOME']
        else:
            jetpack_home = self._get_tmp_home()

        if self.platform == 'nt':
            self.scriptdir = "%s\\config\\healthcheck.d" % jetpack_home
            self.logfile = "%s\\logs\\healthcheck.log" % jetpack_home
            self.sandbox = "%s\\config\\healthcheck.d\\scratch" % jetpack_home
            self.jetpack_cmd = "%s\\bin\jetpack.cmd" % jetpack_home
        else:
            self.scriptdir = "%s/config/healthcheck.d" % jetpack_home
            self.pidfile =  '%s/system/run/healthcheck.pid' % jetpack_home
            self.logfile = '%s/logs/healthcheck.log' % jetpack_home
            self.jetpack_cmd = '%s/bin/jetpack' % jetpack_home
            self.shutdown_proc = self.allocate_shell('/bin/bash')
            self.sandbox = '%s/scratch' % self.scriptdir
            self.warn_proc = self.allocate_shell('wall')
            self.mkfifo()

        self.override_with_cmdline_opts(cmdline_options)

        if not os.path.exists(self.scriptdir):
            logger.fatal("script directory {0} does not exist, bailing out".format(self.scriptdir))
            self.cleanup()
            os.sys.exit(1)
