import click
from jetpack import cli
import logging
import json
import sys
import traceback

logger = logging.getLogger(__name__)

from jetpack.status import JetpackStatus
import jetpack.util
import jetpack.util.cloudinit

@click.group()
@click.option('--loglevel', type=click.Choice(["debug", "info", "warn"]), default='info')
def admin_group(loglevel):
    global logger
    jetpack.util.setup_logging(loglevel)
    logger = logging.getLogger(__name__)


@admin_group.command(name='configure', help="Configures jetpack")
@click.option('--user-data', type=click.Path(), help="Path to the user data file on the node")
@click.option('--jetpack-config', type=click.Path(), help="Path to the jetpack config file to write")
def configure_command(user_data, jetpack_config):
    from jetpack.admin import configure
    try:
        configure.execute(user_data, jetpack_config)
    except configure.JetpackConfigureError as e:
        raise click.ClickException(str(e))
    except Exception:
        raise click.ClickException("An unknown error occurred")

@admin_group.command(name='initialize', help="Initializes the Jetpack installation")
@click.option('--connection', help="Encoded connection string to communicate with CycleCloud")
@click.option('--connection-file', help="JSON formatted connection file to communicate with CycleCloud")
@click.option('--cloud-init', is_flag=True, help="Wait for cloud init to finish before starting Jetpack services")
def initialize_command(connection, connection_file, cloud_init):
    from jetpack import admin

    result = {}
    exception_result = {}

    try:
        admin.initialize(connection, connection_file, cloud_init)
        admin.start_service()
        result["status"] = JetpackStatus.Ok
    except jetpack.util.UserError as e:
        logger.exception("Error installing Jetpack")
        exception_result["status"] = JetpackStatus.Error
        exception_result["message"] = str(e)
        exception_result["description"] = e.description or str(e)
        exception_result["detail"] = e.detail
        exception_result["recommendation"] = e.recommendation
        exception_result["help_link"] = e.help_link
    except Exception:
        logger.exception("Unknown error installing Jetpack")
        exception_result["status"] = JetpackStatus.Error
        exception_result["description"] = "Unexpected error installing Jetpack"
        exception_result["unhandled"] = True
        exception_result["detail"] = traceback.format_exc()
    finally:
        default_recommendation = "Check jetpack.log for more information"

        if exception_result:
            # use the caught exception
            result = exception_result

        # this is used to signal we should include default error information
        if result.get("unhandled"):
            result["recommendation"] = default_recommendation
        
        if result.get("status") == JetpackStatus.Error and "message" not in result:
            result["message"] = "Could not install Jetpack"

        # clean up any attributes used to communicate from the child process (eg unhandled),
        # as well as null values
        valid = {"status", "message", "description", "detail", "recommendation", "help_link"}
        result = {k: v for k, v in result.items() if k in valid and v != None}
    
    sys.stdout.write(json.dumps(result))
    sys.stdout.write("\n")

    if result.get("status", "").lower() != "ok":
        sys.exit(1)


if __name__ == "__main__":
    admin_group()
