import logging
import os
import time

import jetpack.admin
import jetpack.util
import jetpack.users
import jetpack.settings
import jetpack.config
from jetpack.scheduled_event_monitor import ScheduledEventMonitor

# Set up logging that can appropriately detach from the main process with the daemon
logger = logging.getLogger(__name__)


def initialize():
    """ Initialize the service, perform any one time actions that aren't in the main thread loop """

    # Wait for the instance to be registered with CycleCloud - this can take some time due to overprovisioning
    attempts = 0
    while True:
        try:
            config = jetpack.util.get_current_config()
            node_config = config["cyclecloud"]["node"]
            if node_config["ready"]:
                break
            if attempts == 0:
                logger.info("Waiting for overprovisioning to finish")
            elif attempts % 10 == 0:
                logger.info("Overprovisioning not finished yet, waiting...")
            time.sleep(1)
            attempts += 1
        except Exception:
            logger.exception("Failed to get instance status from CycleCloud")

    # By default we enable event monitoring, users can opt out using the new `cyclecloud.monitor_scheduled_events` option
    # or the legacy `cyclecloud.monitor_spot_eviction` option
    if jetpack.config.get('cyclecloud.monitor_scheduled_events', True) and jetpack.config.get('cyclecloud.monitor_spot_eviction', True):
        monitor = ScheduledEventMonitor()
        monitor.start()


def main():
    """ Does the main processing loop we run once a minute or so, other threads may spawn off for custom
    triggers (healthcheck, preemption).

    NOTE: This is meant to be called within an infinite loop
    """
    # Only run the main logic if we have finished our first converge run otherwise just wait for it
    if jetpack.util.has_converged():
        try:
            jetpack.users.manage()
        except Exception as e:
            logger.exception("Uncaught exception: %s" % str(e))
        time.sleep(60)
    else:
        # We should run the converge process, assuming that jetpack has been successfully 
        # installed and that the node is ready (not going to be thrown away by over provisioning)
        # TODO: Util method?
        # NOTE: Jetpack converge has a built in "ready" check, this will be removed in the future
        # when jetpack is part of the install. We will instead have to check for "ready" here and 
        # call appropriate onReady scripts to start services and what not
        if os.path.exists(os.path.join(jetpack.util.jetpack_home(), 'run', 'jetpack.installed')):            
            try:
                import subprocess
                jetpack.util.set_cyclecloud_env()
                # NOTE: We fork a converge here instead of jetpack.converge.execute() because we
                # don't want to screw up the log files
                logger.info("Forking `jetpack converge` to configure VM software")
                subprocess.check_call('jetpack converge --mode=install', shell=True)
                logger.info("Software configuration complete!")
            except:
                logger.exception("Failed to configure software")
        else:
            logger.info("Waiting for Jetpack installation to complete before configuring software...")
        time.sleep(10)
