import os
import logging
import json
from jetpack import util
import jetpack.converge.userdata

logger = logging.getLogger()

UNKNOWN_ERROR = "An unknown error occurred while processing the settings data file."

SETTINGS_FILE = os.path.join(util.jetpack_home(), 'config', 'settings.json')

class SettingsError(Exception):
    pass

def get(value=None, default=None):
    try:
        current_val = _get_data()
        if value is None:
            return current_val
        properties = value.split('.')
        for prop in properties:
            current_val = current_val[str(prop)]
        return current_val
    except KeyError:
        if default is not None:
            return default
        else:
            raise SettingsError("The value '%s' was not found in the node settings." % value)
    except Exception:
        logger.exception("")
        raise SettingsError(UNKNOWN_ERROR)


def update(data):
    if not os.path.isfile(SETTINGS_FILE):
        with open(SETTINGS_FILE, 'w') as f:
            f.write(json.dumps(data, indent=4, sort_keys=True))


def available():
    """ Determines if the settings are available to be queried """
    return os.path.isfile(SETTINGS_FILE)


def _get_data():
    """ Read the settings data from the cached file, otherwise pull it out of userdata """
    if not os.path.isfile(SETTINGS_FILE):
        return jetpack.converge.userdata.get_settings()
    try:
        with open(SETTINGS_FILE) as f:
            settings = json.loads(f.read())
            return settings
    except IOError as e:
        raise SettingsError(e.strerror)
    except ValueError as e:
        raise SettingsError(str(e))
    except Exception:
        logger.exception("Unhandled exception getting settings data")
        raise SettingsError(UNKNOWN_ERROR)
