import json
import os
import subprocess
import time

import jetpack.util

CLOUD_INIT_BOOT_FILE = "/var/lib/cloud/instance/boot-finished"
CLOUD_INIT_RESULT_FILE = "/var/lib/cloud/data/result.json"

class CloudInitException(Exception):
    pass

def installed():
    """ Check to see if cloud-init is available """
    if not jetpack.util.is_linux():
        return False
    try:
        subprocess.check_output("which cloud-init", shell=True, stderr=subprocess.STDOUT)
        return True
    except subprocess.CalledProcessError:
        return False

def enabled():
    """ Check to see if cloud-init is enabled: https://cloudinit.readthedocs.io/en/latest/topics/boot.html"""
    enabled = installed()

    # The most basic way to disable cloud-init is to touch /etc/cloud/cloud-init.disabled
    if os.path.isfile('/etc/cloud/cloud-init.disabled'):
        enabled = False

    # Or change the kernel command line to disabled it    
    if os.path.isfile('/proc/cmdline'):
        with open('/proc/cmdline') as f:
            if "cloud-init=disabled" in f.read():
                enabled = False

    # In a container, $KERNEL_CMDLINE should be set and we can check for cloud-init=disabled there
    if "cloud-init=disabled" in os.environ.get('KERNEL_CMDLINE',''):
        enabled = False
        
    return enabled

def finished():
    """ Check to see if cloud-init has finished running (success/fail doesn't matter) """

    # NOTE: In newer versions of cloud-init we can run `cloud-init status` and parse the output

    # Cloud-init touches a file when it is finished running regardless of success/fail
    return os.path.isfile(CLOUD_INIT_BOOT_FILE)

def wait(timeout=3600, delay=5):
    """ Wait for the cloud-init to finish running and return the errors if any exist """

    # NOTE: In newer versions of cloud-init we can just run `cloud-init status --wait`
    while not finished():
        time.sleep(delay)
        timeout -= delay
        if timeout <= 0:
            raise jetpack.util.UserError("Failed to run Cloud-Init", description="Timeout expired while waiting for cloud-init to finish running")

    return errors()

def errors():
    """ Get the errors from the cloud-init run, empty list if there were no errors """
    # Read results.json from disk and return the error set if it exists
    errors = []
    if os.path.isfile(CLOUD_INIT_RESULT_FILE):
        result = {}
        with open(CLOUD_INIT_RESULT_FILE) as f:
            result = json.load(f)
        errors = result.get("v1", {}).get("errors", [])

    return errors

