"""
This file assists with invoking commands in a way that the admin can call it directly and get the exact same behavior.
The contents of the command are written to a temporary script file, including explicitly setting all environment variables,
and then the script is executed. This is useful for the kind of "system operations" that, once we have tested them,
only fail in the customer's environment due to their specific operating system, networking settings, or permissions.
"""

from abc import ABC, abstractmethod
from io import StringIO
import shlex
import os
import stat
import subprocess
import tempfile
import time
import socket
from contextlib import contextmanager
from typing import List, Optional
import logging
import jetpack.util


# how many command files to keep around regardless of age
MAX_FILES: int = 100
# how long to keep old command files around (TODO: we may not need this since we have the max-file limit)
OLD_FILE_CUTOFF: int = 86400

logger = logging.getLogger(__name__)


class CommandExecution:
    """A wrapper around a subprocess.Popen object that provides a recommendation for debugging"""
    def __init__(self, proc, script):
        self.proc = proc
        self.script = script

    def get_recommendation(self):
        ip_addr = socket.gethostbyname(socket.gethostname())
        return f"Please run the following command on {ip_addr} to debug further:\n{self.script}"


@contextmanager
def begin(name: str, cmd: list, env: Optional[dict] = None, cwd: Optional[str] = None):
    """Opens a subprocess and returns a CommandExecution object that contains the Popen object.
    The script is not deleted, but only cleaned up on subsequent runs."""

    script_name = runner.create_script(name, cmd, env, cwd)

    with subprocess.Popen(script_name, stdout=subprocess.PIPE, text=True) as proc:
        yield CommandExecution(proc, script_name)


class CommandRunner(ABC):
    """Base class for code that knows how to write a script to run a command"""

    # the cached string of export FOO=bar for each environment variable.
    # This assumes that the environment variables don't change during the lifetime of the process.
    cached_script_prefix: Optional[str] = None

    def create_script(
        self,
        name: str,
        cmd: list,
        env: Optional[dict] = None,
        cwd: Optional[str] = None,
    ):
        '''Writes a script file that runs the given command'''
        command_dir = os.path.join(jetpack.util.jetpack_home(), "run/commands")

        prepare_dir(command_dir)

        with tempfile.NamedTemporaryFile(
            dir=command_dir, prefix=f"{name}_", suffix=self.get_shell_ext(), delete=False, mode="w"
        ) as f:
            script_name = os.path.abspath(f.name)
            f.write(self.get_script_contents(cmd, env, cwd))

        st = os.stat(script_name)
        os.chmod(script_name, st.st_mode | stat.S_IEXEC)

        return script_name

    def get_script_contents(
        self,
        cmd: list,
        env: Optional[dict] = None,
        cwd: Optional[str] = None,
    ) -> str:
        '''Returns the contents of the script file that runs the given command'''
        s = StringIO()
        
        if CommandRunner.cached_script_prefix is None:
            CommandRunner.cached_script_prefix = self.create_prefix()
        s.write(CommandRunner.cached_script_prefix)

        # write out the custom environment variables
        if env:
            s.write(self.write_env_vars(env))

        s.write("\n")

        # include an explicit cd to the working directory for the command
        cwd = cwd if cwd else os.curdir
        s.write(f"cd {os.path.abspath(cwd)}\n")

        # write out the command
        s.write(self.get_invocation(cmd))

        s.write("\n\n")

        return s.getvalue()

    def write_env_vars(self, env_vars) -> str:
        """Returns the string to set the given environment variables in the script"""
        result = ""
        for k, v in env_vars.items():
            result += self.get_env_var(k, v)
        return result

    @abstractmethod
    def create_prefix(self) -> str:
        '''Returns the prefix of the script, which includes the static environment variables'''

    @abstractmethod
    def get_comment(self, comment: str) -> str:
        """Returns the string to comment out the given text in the script"""

    @abstractmethod
    def get_env_var(self, name: str, value: str) -> str:
        """Returns the string to set the given environment variable in the script"""

    @abstractmethod
    def get_invocation(self, cmd: List[str]) -> str:
        """Returns the string to invoke the given command in the script"""

    @abstractmethod
    def get_shell_ext(self) -> str:
        """Returns the file extension for the script file"""


class LinuxCommandRunner(CommandRunner):
    """Linux implementation of CommandRunner"""

    def get_comment(self, comment: str):
        return f"# {comment}"

    def get_env_var(self, name: str, value: str):
        # On some distros, bash functions end up in the environment (eg `BASH_FUNC_which_%%`), which can't be exported
        if not name.startswith("BASH_FUNC_"):
            return f"export {name}={shlex.quote(value)}\n"
        return ""

    def get_invocation(self, cmd: List[str]):
        quoted = " ".join([shlex.quote(c) for c in cmd])
        return f"{quoted}"

    def get_shell_ext(self) -> str:
        return ".sh"
    
    def create_prefix(self):
        script_prefix = """#!/bin/bash

# re-run with no environment variables except the bare minimum (PWD, OLDPWD, SHLVL)
if [ "$USER" != "" ]; then env -i bash --noprofile --norc "$BASH_SOURCE" ; exit ; fi

# Jetpack environment variables:
"""
        script_prefix += self.write_env_vars(os.environ)
        script_prefix += "\n"
        script_prefix += self.get_comment(
            "Custom environment variables for this command:\n"
        )
        return script_prefix

class WindowsCommandRunner(CommandRunner):
    """Windows implementation of CommandRunner"""

    def create_prefix(self) -> str:
        script_prefix = """@ECHO OFF

REM clear existing environment variables
FOR /f "tokens=1,* delims==" %%f in ('set') do SET "%%f="

REM Jetpack environment variables:
"""
        script_prefix += self.write_env_vars(os.environ)
        script_prefix += "\n"

        script_prefix += self.get_comment(
            "Custom environment variables for this command:\n"
        )

        return script_prefix

    def get_comment(self, comment: str):
        return f"REM {comment}"

    def get_env_var(self, name: str, value: str):
        if not "(" in name:
            # the batch syntax is SET "NAME=VALUE" and it doesn't matter if there are quotes in the value
            return f'SET "{name}={value}"\n'
        return ""

    def get_invocation(self, cmd: List[str]):
        quoted = " ".join([self._quote(c) for c in cmd])
        return f"{quoted}"

    def get_shell_ext(self) -> str:
        return ".cmd"

    def _quote(self, value):
        # inside double quotes the only thing we need to care about is percent signs and more double quotes
        value = value.replace('"', '""')
        value = value.replace('%', '%%')
        return f'"{value}"'


runner: CommandRunner = WindowsCommandRunner() if jetpack.util.is_windows() else LinuxCommandRunner()


def prepare_dir(
    command_dir: str,
    now: float = time.time(),
    cutoff: int = OLD_FILE_CUTOFF,
    max_files: int = MAX_FILES,
):
    """Creates the command directory and cleans up old files"""
    os.makedirs(command_dir, exist_ok=True)

    old_file_cutoff = cutoff

    retained: list = []
    for f in os.listdir(command_dir):
        filename = os.path.join(command_dir, f)
        if os.path.isfile(filename):
            if os.path.getmtime(filename) < now - old_file_cutoff:
                os.remove(filename)
            else:
                retained.append({"name": f, "mtime": os.path.getmtime(filename)})

    if len(retained) > max_files:
        retained.sort(key=lambda x: x["mtime"])
        for item in retained[: len(retained) - max_files]:
            os.remove(os.path.join(command_dir, item["name"]))
