import unittest
from unittest import mock
import json
from jetpack import cluster_init
import jetpack.util
from jetpack.lockers import Locker
from subprocess import STDOUT
import os

from jetpack.status import JetpackResource, JetpackStatus

class TestClusterInit(unittest.TestCase):

    def setUp(self):
        self.spec = cluster_init.ClusterInitSpec({
            'name': 'mytestspec',
            'project': 'test-project',
            'project_spec': 'test-spec',
            'version': '1.0.0',
            'location': 's3://com.cyclecomputing.test/projects/test-project/test-spec/1.0.0',
            'creds': {
                'access_key': 'my-key',
                'secret_key': 'my-secret'
            }
        })

    @mock.patch('os.path.isdir', return_value=False)
    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('jetpack.cluster_init.ClusterInitSpec._download')
    @mock.patch('os.makedirs')
    def test_load_from_manifest(self, mock_makedirs, mock_download, mock_isfile, mock_isdir):

        with mock.patch('jetpack.cluster_init.open', mock.mock_open(), create=True) as mock_open:
            mock_open.return_value.read.return_value = json.dumps({
                "specs": {
                    "test-spec": {
                        "run_list": ["a", "b", "c"],
                        "other": "value"
                    },
                    "other-spec": {
                        'run_list': ["nada"]
                    }
                }
            })
            spec = cluster_init.ClusterInitSpec({
                'name': 'test-spec',
                'project': 'test-project',
                'project_spec': 'test-spec',
                'version': '1.0.0',
                'location': 'projects/test-project/1.0.0/test-spec',
                'creds': {
                    'access_key': 'access',
                    'secret_key': 'secret'
                }
            }, load_from_manifest=True)

            mock_open.assert_called_once_with('/opt/cycle/jetpack/config/projects/test-project/manifest.json', 'r')

            mock_makedirs.assert_called_once_with('/opt/cycle/jetpack/config/projects/test-project')
            mock_download.assert_called_once_with('projects/test-project/1.0.0/manifest.json', '/opt/cycle/jetpack/config/projects/test-project/manifest.json', verify_source=True)
            mock_isfile.assert_called_once_with('/opt/cycle/jetpack/config/projects/test-project/manifest.json')
            self.assertEqual(["a", "b", "c"], spec.run_list)

            # Now test to make sure users can override the runlist
            spec = cluster_init.ClusterInitSpec({
                'name': 'test-spec',
                'project': 'test-project',
                'project_spec': 'test-spec',
                'version': '1.0.0',
                'location': 'projects/test-project/1.0.0/test-spec',
                'creds': {
                    'access_key': 'access',
                    'secret_key': 'secret'
                },
                "run_list": ["user", "defined", "runlist"]
            }, load_from_manifest=True)

            self.assertEqual(["user", "defined", "runlist"], spec.run_list)


    @mock.patch('jetpack.cluster_init.ClusterInitSpec._load_from_manifest')
    @mock.patch('jetpack.config.get', return_value=[])
    @mock.patch('jetpack.cluster_init._get_cluster_init_home', return_value='/mnt/cluster-init')
    def test_get_specs(self, mock_get_cl_home, mock_config_get,
                       mock_load_from_manifest):

        # No specs should be an empty list
        mock_config_get.return_value = []
        self.assertEqual([], cluster_init.get_specs())

        # Fully formed userdata should get us a list of spec dicts
        mock_config_get.return_value = [
            {"name": "foo", "project": "project1", "version": "1.0.0", "project_spec": "spec1", "locker": "test-locker", "location": "path/test-1"},
            {"name": "bar", "project": "project2", "version": "2.0.0", "project_spec": "spec2", "locker": "test-locker", "location": "path/test-2"}
        ]

        specs = cluster_init.get_specs()
        self.assertEqual(2, len(specs))
        self.assertEqual("project1", specs[0].project_name)
        self.assertEqual("spec1", specs[0].name)
        self.assertEqual("1.0.0", specs[0].project_version)
        self.assertEqual("/mnt/cluster-init/project1/spec1", specs[0].path)
        self.assertEqual("path/test-1", specs[0].location)

        self.assertEqual("project2", specs[1].project_name)
        self.assertEqual("spec2", specs[1].name)
        self.assertEqual("2.0.0", specs[1].project_version)

        # # Test filtering by project
        specs = cluster_init.get_specs(project_name="project1")
        self.assertEqual(1, len(specs))
        self.assertEqual("project1", specs[0].project_name)

        # # Exception should be raised if it's an incomplete specs
        mock_config_get.return_value = [
            {"name": "foo", "project_spec": "spec3"}
        ]
        self.assertRaises(cluster_init.ClusterInitError, cluster_init.get_specs)

        # # Ensure that 'cyclecloud' project is filtered by default
        mock_config_get.return_value = [
            {"name": "cyclecloud:default:7.9.0", "project": "cyclecloud", "version": "7.9.0", "project_spec": "default"}
        ]
        specs = cluster_init.get_specs()
        self.assertEqual(0, len(specs))

        mock_config_get.return_value = [
            {"name": "foo", "project": "project4", "version": "4.0.0", "project_spec": "spec4"}
        ]
        specs = cluster_init.get_specs()
        self.assertEqual(1, len(specs))
        self.assertEqual("project4", specs[0].project_name)

    @mock.patch('jetpack.lockers.get')
    @mock.patch('jetpack.util.azcopy.sync')
    @mock.patch('jetpack.util.azcopy.copy')
    @mock.patch('jetpack.util.azcopy.file_exists')
    @mock.patch('os.path.isfile')
    def test_download(self, mock_isfile, mock_file_exists, mock_copy, mock_sync, mock_lockers_get):
        locker = mock.Mock(spec=Locker)
        locker.url.return_value = "https://account.blob.core.windows.net/path/to/file.tgz"
        mock_lockers_get.return_value = locker
        
        spec = cluster_init.ClusterInitSpec({
            'name': 'mytestspec',
            'project': 'test-project',
            'project_spec': 'test-spec',
            'version': '1.0.0',
            'creds': {
                'environment': {'storage_suffix': '.core.windows.net'},
                'sas_token': '?TEST_SAS_TOKEN',
            }
        })

        # Single file (new file)
        mock_isfile.return_value = False
        spec._download("path/to/file.tgz", "/tmp")
        mock_isfile.assert_called_once_with('/tmp')
        mock_copy.assert_called_once_with('path/to/file.tgz', '/tmp', src_locker=locker)

        # Single file (already exists)
        mock_isfile.reset_mock()
        mock_isfile.return_value = True
        spec._download("path/to/file.tgz", "/tmp")
        mock_isfile.assert_called_once_with('/tmp')
        mock_sync.assert_called_once_with('path/to/file.tgz', '/tmp', src_locker=locker, recursive=True, delete_destination=False)

        # Single file, verify source (success)
        mock_isfile.reset_mock()
        mock_sync.reset_mock()
        mock_isfile.return_value = True
        mock_file_exists.return_value = True
        spec._download("path/to/file.tgz", "/tmp", verify_source=True)
        mock_isfile.assert_called_once_with('/tmp')
        mock_sync.assert_called_once_with('path/to/file.tgz', '/tmp', src_locker=locker, recursive=True, delete_destination=False)

        # Single file, verify source (failure)
        mock_file_exists.return_value = False
        with self.assertRaises(cluster_init.SourceNotFound):
            spec._download("path/to/file.tgz", "/tmp", verify_source=True)

        # Directory
        mock_isfile.reset_mock()
        mock_sync.reset_mock()
        spec._download("path/to/directory", "/tmp", directory=True)
        self.assertFalse(mock_isfile.called)
        mock_sync.assert_called_once_with('path/to/directory/', '/tmp/', src_locker=locker, recursive=True, delete_destination=False)

        # Windows Directory
        mock_isfile.reset_mock()
        mock_sync.reset_mock()
        spec._download("path/to/directory", "c:\\tmp", directory=True)
        self.assertFalse(mock_isfile.called)
        mock_sync.assert_called_once_with('path/to/directory/', 'c:\\tmp/', src_locker=locker, recursive=True, delete_destination=False)


    @mock.patch('jetpack.cluster_init.ClusterInitSpec._download')
    @mock.patch("jetpack.config.get")
    def test_download_scripts(self, mock_config_get, download):
        mock_config_get.return_value = "/tmp/cluster-init"
        
        # Validate that we figure out the paths correctly
        self.spec.download_scripts()
        download.assert_called_once_with('s3://com.cyclecomputing.test/projects/test-project/test-spec/1.0.0/cluster-init', '/tmp/cluster-init/test-project/test-spec', directory=True)
    
    @mock.patch('jetpack.cluster_init.ClusterInitSpec._has_windows_line_endings')
    @mock.patch('jetpack.cluster_init.Popen')
    @mock.patch('jetpack.util.is_windows', return_value=True)
    def test_run_script(self, mock_is_windows, mock_popen, mock_has_windows_line_endings):
        with mock.patch('jetpack.cluster_init.open', mock.mock_open(), create=True) as mock_open:

            # Test windows
            self.spec._run_script("test.sh", env={'test': '123'}, cwd='/tmp', output_file="output.txt")
            mock_is_windows.assert_called_once()
            mock_popen.assert_called_once_with('test.sh', env={'test': '123'}, cwd='/tmp', stderr=STDOUT, stdout=mock_open.return_value.__enter__.return_value, shell=True)

            # Test linux
            mock_popen.reset_mock()
            mock_is_windows.return_value = False
            self.spec._run_script("test.sh", env={'test': '123'}, cwd='/tmp', output_file="output.txt")
            mock_popen.assert_called_once_with(['test.sh'], env={'test': '123'}, cwd='/tmp', stderr=STDOUT, stdout=mock_open.return_value.__enter__.return_value)

            # Test OSError
            mock_popen.reset_mock()
            mock_popen.side_effect = OSError(2, "File not found")
            with self.assertRaises(jetpack.util.UserError) as e:
                self.spec._run_script("test.sh", env={'test': '123'}, cwd='/tmp', output_file="output.txt")
                mock_has_windows_line_endings.assert_called_once_with('test.sh')
                self.assertTrue(False)

    @mock.patch('os.environ.copy')
    @mock.patch('os.listdir')
    @mock.patch('jetpack.cluster_init.ClusterInitSpec._run_script')
    def test_run(self, mock_run_script, mock_listdir, mock_copy):
        
        # Mock out a bunch of the stuff that we probably don't need to call directly
        cluster_init._make_executable = mock.Mock()
        self.spec._create_directories = mock.Mock()
        self.spec._create_directories.return_value = ("/tmp/run", "/tmp/scripts", "/tmp/logs")
        mock_copy.return_value = {'INHERITED': 'FROM OS ENVIRON'}
        
        # listdir returns files in a random order, order them as such...
        mock_listdir.return_value = ['03_test.sh', 'foo.sh', '01_test2.sh']

        # Return value of the process should be 0 (success)
        mock_run_script.return_value.returncode = 0

        with mock.patch('jetpack.cluster_init.open', mock.mock_open(), create=True) as mock_open:
            self.spec.run(root="/tmp/cluster-init", fail_on_error=True, allowed_executables=[".sh"], rerun=False)

            env = {
                'INHERITED': 'FROM OS ENVIRON',
                'CYCLECLOUD_PROJECT_NAME': 'test-project', 
                'CYCLECLOUD_PROJECT_PATH': '/tmp/cluster-init/test-project', 
                'CYCLECLOUD_SPEC_NAME': 'test-spec', 
                'CYCLECLOUD_SPEC_PATH': '/tmp/cluster-init/test-project/test-spec'
            }

            # The calls should be made in the correct order
            mock_run_script.assert_has_calls([
                mock.call('/tmp/scripts/01_test2.sh', env, '/tmp/scripts', '/tmp/logs/01_test2.sh.out'),
                mock.call('/tmp/scripts/03_test.sh', env, '/tmp/scripts', '/tmp/logs/03_test.sh.out'),
                mock.call('/tmp/scripts/foo.sh', env, '/tmp/scripts', '/tmp/logs/foo.sh.out')
            ])
            self.assertEqual(3, mock_open.call_count)

        # Now with something not in the include list
        mock_listdir.return_value = ['test1.sh', 'other.txt']
        mock_run_script.reset_mock()

        with mock.patch('jetpack.cluster_init.open', mock.mock_open(), create=True) as mock_open:
            self.spec.run(root="/tmp/cluster-init", fail_on_error=True, allowed_executables=[".sh"], rerun=False)

            mock_run_script.assert_called_once_with('/tmp/scripts/test1.sh', env, '/tmp/scripts', '/tmp/logs/test1.sh.out')
            self.assertEqual(1, mock_open.call_count)

        # Now if there is a script error, we should bail early and raise an exception
        mock_run_script.reset_mock()
        mock_run_script.return_value.returncode = 1

        with mock.patch('jetpack.cluster_init.open', mock.mock_open(), create=True) as mock_open:
            self.assertRaises(jetpack.util.UserError, self.spec.run, root="/tmp/cluster-init", fail_on_error=True, allowed_executables=[".sh"], rerun=False)

    @mock.patch('os.makedirs')
    @mock.patch('jetpack.cluster_init.os.path.isdir', return_value=False)
    def test_create_directories(self, mock_isdir, mock_makedirs):
        mock_isdir.return_value = False
        
        run_dir, scripts_dir, log_dir = self.spec._create_directories(root='/tmp')

        self.assertEqual(3, mock_isdir.call_count)
        self.assertEqual(3, mock_makedirs.call_count)

        self.assertEqual(run_dir, "/tmp/.run/test-project/test-spec/scripts")
        self.assertEqual(scripts_dir, "/tmp/test-project/test-spec/scripts")
        self.assertEqual(log_dir, "/opt/cycle/jetpack/logs/cluster-init/test-project/test-spec/scripts")

    @mock.patch('jetpack.cluster_init.get_specs')
    @mock.patch('jetpack.status.update')
    @mock.patch('jetpack.status.send')
    def test_process_scripts(self, mock_send, mock_update, get_specs):
        spec1 = mock.Mock()
        spec2 = mock.Mock()
        get_specs.return_value = [spec1, spec2]

        cluster_init.process_scripts(no_sync=False, rerun=False)

        mock_update.assert_called_once_with(jetpack.status.JetpackResource.SoftwareConfiguration, jetpack.status.JetpackStatus.Ok, "Cluster-init completed successfully")

        spec1.download_scripts.assert_called_once()
        spec2.download_scripts.assert_called_once()
        spec1.run.assert_called_once_with(rerun=False)
        spec2.run.assert_called_once_with(rerun=False)

        # Run again in "dev" mode, no sync, and rerun is true
        spec1.reset_mock()
        spec2.reset_mock()

        cluster_init.process_scripts(no_sync=True, rerun=True)
        self.assertFalse(spec1.download_scripts.called)
        self.assertFalse(spec2.download_scripts.called)
        spec1.run.assert_called_once_with(rerun=True)
        spec2.run.assert_called_once_with(rerun=True)

    @mock.patch('jetpack.cluster_init.get_specs')
    def test_process_chef(self, get_specs):
        spec1 = mock.Mock()
        spec2 = mock.Mock()
        get_specs.return_value = [spec1, spec2]

        cluster_init.process_chef()

        spec1.download_chef.assert_called_once()
        spec2.download_chef.assert_called_once()


    @mock.patch('os.path.exists')
    @mock.patch('os.path.isdir')
    @mock.patch('shutil.rmtree')
    @mock.patch('shutil.move')
    def test_move_and_overwrite(self, mock_move, mock_rmtree, mock_isdir, mock_exists):
        # Case 1: the destination doesn't exist
        # We should just call a shutil.move
        mock_exists.return_value= False
        cluster_init._move_and_overwrite("/this/is/a/directory", "/doesnt/exist/")
        self.assertEqual(0, mock_rmtree.call_count)
        mock_move.assert_called_once_with("/this/is/a/directory", "/doesnt/exist/")

        # Case 2: The destination exists and is NOT a directory
        # We should also just call a shutil.move
        mock_rmtree.reset_mock()
        mock_move.reset_mock()
        mock_exists.return_value = True
        mock_isdir.return_value = False
        cluster_init._move_and_overwrite("/this/is/a/directory", "/doesnt/exist/")
        self.assertEqual(0, mock_rmtree.call_count)
        mock_move.assert_called_once_with("/this/is/a/directory", "/doesnt/exist/")

        # Case 3: The destination exists and IS a directory
        # We have to shutil.rmtree the destination first before calling shutil.move
        mock_rmtree.reset_mock()
        mock_move.reset_mock()
        mock_exists.return_value = True
        mock_isdir.return_value = True
        cluster_init._move_and_overwrite("/this/is/a/directory", "/does/exist/")
        mock_rmtree.assert_called_once_with("/does/exist/")
        mock_move.assert_called_once_with("/this/is/a/directory", "/does/exist/")

        
    @mock.patch('os.makedirs')
    @mock.patch('jetpack.cluster_init.ClusterInitSpec._download')
    @mock.patch('jetpack.util.is_windows', return_value=False)
    @mock.patch('jetpack.cluster_init._process_chef_tarball')
    def test_download_chef(self, mock_process_chef_tarball, mock_is_windows, mock_download, mock_makedirs):

        self.spec.download_chef()
        
        mock_makedirs.assert_called_once_with('/opt/cycle/jetpack/system/chef/tarballs/test-project/test-spec')
        mock_download.assert_called_once_with('s3://com.cyclecomputing.test/projects/test-project/test-spec/1.0.0/chef', '/opt/cycle/jetpack/system/chef/tarballs/test-project/test-spec', delete_removed=True, directory=True)
        
        mock_process_chef_tarball.assert_any_call('/opt/cycle/jetpack/system/chef/tarballs/test-project/test-spec', 'site-cookbooks.tgz', '/opt/cycle/jetpack/system/chef/chef-repo', 'cookbooks')
        mock_process_chef_tarball.assert_any_call('/opt/cycle/jetpack/system/chef/tarballs/test-project/test-spec', 'roles.tgz', '/opt/cycle/jetpack/system/chef/chef-repo', 'roles')
        mock_process_chef_tarball.assert_any_call('/opt/cycle/jetpack/system/chef/tarballs/test-project/test-spec', 'data_bags.tgz', '/opt/cycle/jetpack/system/chef/chef-repo', 'data_bags')

    @mock.patch('os.path.isdir', return_value=False)
    @mock.patch('os.makedirs')
    @mock.patch('shutil.rmtree')
    @mock.patch('tempfile.mkdtemp')
    @mock.patch('jetpack.cluster_init._move_and_overwrite')
    @mock.patch('os.listdir')
    @mock.patch('jetpack.cluster_init.tarfile')
    @mock.patch('os.path.isfile')
    def test_process_chef_tarball(self, mock_isfile, mock_tarfile, mock_listdir, mock_move_and_overwrite, mock_mkdtemp, mock_rmtree, mock_makedirs, mock_isdir):
        mock_isfile.return_value = True
        mock_listdir.return_value = ['test1', 'test2.txt']
        mock_mkdtemp.return_value = "/random/temp"

        cluster_init._process_chef_tarball('/tmp/src', 'site-cookbooks.tgz', '/tmp/dst', 'cookbooks')

        mock_tarfile.open.return_value.__enter__.return_value.extractall.assert_called_once_with('/random/temp')

        self.assertEqual(2, mock_move_and_overwrite.call_count)
        mock_move_and_overwrite.assert_any_call('/random/temp/site-cookbooks/test1', '/tmp/dst/cookbooks/test1')
        mock_move_and_overwrite.assert_any_call('/random/temp/site-cookbooks/test2.txt', '/tmp/dst/cookbooks/test2.txt')
        mock_rmtree.assert_called_once_with("/random/temp")
        mock_makedirs.assert_called_once_with("/tmp/dst/site-cookbooks")

    @mock.patch('jetpack.cluster_init.get_specs')
    @mock.patch('jetpack.cluster_init.ClusterInitSpec._download')
    def test_download_blob(self, mock_download, mock_get_specs):

        spec1 = cluster_init.ClusterInitSpec(spec={
            'location': 'projects/test%2Fproject/1.2.3/test-spec',
            'project': 'test/project',
            'project_spec': 'test-spec',
            'version': '1.2.3'
        })
        spec2 = cluster_init.ClusterInitSpec(spec={
            'location': 'projects/test-project/1.2.3/other-spec',
            'project': 'test-project',
            'project_spec': 'other-spec',
            'version': '1.2.3.'
        })

        mock_get_specs.return_value = [spec1, spec2]
        cluster_init.download_blob("path/to/file.tgz", "/tmp/destination", "test/project", "test-spec")
        mock_get_specs.assert_called_once_with(project_name="test/project", name="test-spec")
        mock_download.assert_called_once_with("blobs/test%2Fproject/path/to/file.tgz", "/tmp/destination", verify_source=True)

        # Test cascading download, totally mock out the _download command so we can control what things are thrown
        mock_download.reset_mock()
        def _download(src, dst, command="sync", options=None, verify_source=False):
            if src == "blobs/test%2Fproject/path/to/user/defined/file.tgz":
                raise cluster_init.SourceNotFound("Oh noes not found!")
            if src == "blobs/cached-project/path/to/user/defined/file.tgz":
                raise cluster_init.SourceNotFound("Yet another file not found!")
        mock_download.side_effect = _download

        cluster_init.download_blob("path/to/user/defined/file.tgz", "/tmp/", "test/project", "test-spec")
        
        # We should get two download calls here, the first is to the "project blobs" which raises a SourceNotFound
        # After which we will attempt to download from the user defined blobs
        mock_download.assert_has_calls([
            mock.call('blobs/test%2Fproject/path/to/user/defined/file.tgz', '/tmp/', verify_source=True),
            mock.call('projects/test%2Fproject/blobs/path/to/user/defined/file.tgz', '/tmp/', verify_source=True)
        ])

        # Test to ensure everything works as expected for a 'cached' project
        mock_download.reset_mock()
        spec1 = cluster_init.ClusterInitSpec(spec={
            'location': 'cache/projects/cached-project/1.2.3/cached-spec',
            'project': 'cached-project',
            'project_spec': 'cached-spec',
            'version': '1.2.3'
        })
        mock_get_specs.return_value = [spec1]

        cluster_init.download_blob("path/to/user/defined/file.tgz", "/tmp/", "cached-project", "cached-spec")
        
        # We should get two download calls here, the first is to the "project blobs" which raises a SourceNotFound
        # After which we will attempt to download from the user defined blobs
        mock_download.assert_has_calls([
                        mock.call('blobs/cached-project/path/to/user/defined/file.tgz', '/tmp/', verify_source=True),
                        mock.call('cache/projects/cached-project/blobs/path/to/user/defined/file.tgz', '/tmp/', verify_source=True)
        ])
